import json
from pathlib import Path
from typing import Any, Dict, List, Optional
from urllib.parse import urlparse

import yaml
from attr import define
from email_validator import EmailNotValidError
from email_validator import (
    ValidatedEmail as _ValidatedEmail,
)
from email_validator import validate_email as _validate_email
from pydantic import AfterValidator, BaseModel, ConfigDict, Field, SecretStr
from typing_extensions import Annotated


def validate_version(value: str) -> str:
    if value != "v1":
        raise ValueError("Version unsupported")
    return value


def validate_url(value: str) -> str:
    result = urlparse(value)
    assert all(field for field in (result.scheme, result.netloc, result.path))
    return value


def validate_email(
    value: Optional[str], *, raise_on_error: bool = True
) -> Optional[str]:
    if value is None:
        return None
    try:
        validated_email: _ValidatedEmail = _validate_email(
            value, check_deliverability=False
        )
        return validated_email.normalized
    except EmailNotValidError as e:
        if raise_on_error:
            raise e
        else:
            return None


ValidatedVersion = Annotated[str, AfterValidator(validate_version)]
ValidatedURL = Annotated[str, AfterValidator(validate_url)]
ValidatedEmail = Annotated[str, AfterValidator(validate_email)]


@define(frozen=True)
class Defaults:
    desktop_useragent = (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
        "AppleWebKit/537.36 (KHTML, like Gecko) "
        "Chrome/115.0.0.0 Safari/537.36 Edg/115.0.1901.188"
    )
    mobile_useragent = (
        "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 "
        "(KHTML, like Gecko) Chrome/115.0.0.0 Mobile Safari/537.3"
    )

    rewards_homepage = "https://rewards.bing.com/"
    bing_homepage = "https://www.bing.com/?scope=web"


class Profile(BaseModel):
    email: ValidatedEmail
    profile: str
    skip: bool = False


class AutomsrConfig(BaseModel):
    """
    >>> profiles = [{"email": "1@gmail.com", "profile": "p1"}, {"email": "2@gmail.com", "profile": "p2"}]
    >>> _ = AutomsrConfig(profiles=profiles)
    """

    profiles: List[Profile] = Field(..., min_length=1)

    rewards_homepage: ValidatedURL = Defaults.rewards_homepage
    bing_homepage: ValidatedURL = Defaults.bing_homepage

    desktop_useragent: str = Defaults.desktop_useragent
    mobile_useragent: str = Defaults.mobile_useragent


class EmailConfig(BaseModel):
    enable: bool = False
    recipient: Optional[ValidatedEmail] = None
    sender: Optional[ValidatedEmail] = None
    sender_password: Optional[SecretStr] = None
    host: Optional[str] = None
    port: Optional[int] = None
    tls: Optional[bool] = False


class SeleniumConfig(BaseModel):
    profiles_root: Path
    chromedriver_path: Path


class Config(BaseModel):
    model_config = ConfigDict(extra="forbid")

    version: ValidatedVersion = "v1"
    automsr: AutomsrConfig
    email: EmailConfig
    selenium: SeleniumConfig

    @classmethod
    def from_yaml(cls, path: Path) -> "Config":
        """
        Load config from a yaml filepath.
        """

        data = yaml.safe_load(open(path))
        return cls.from_dict(data)

    @classmethod
    def from_json(cls, path: Path) -> "Config":
        """
        Load config from a json filepath.
        """

        data = json.load(open(path))
        return cls.from_dict(data)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Config":
        """
        Load config from an in-memory mapping.

        >>> profiles = [{"email": "1@gmail.com", "profile": "p1"}]
        >>> _data = {
        ...     "automsr": {"profiles": profiles},
        ...     "email": {},
        ...     "selenium": {"profiles_root": Path("profiles"), "chromedriver_path": Path("chromedriver.exe")},
        ... }
        >>> Config.from_dict(_data)  # doctest: +ELLIPSIS
        Config(version='v1', automsr=..., email=..., selenium=...)
        >>> _data["email"]["sender"] = "invalid-email"
        >>> Config.from_dict(_data)  # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        pydantic_core._pydantic_core.ValidationError: 1 validation error for Config
        email.sender
          Value error, The email address is not valid...
        ...
        >>> del _data["email"]["sender"]
        >>> _data["version"] = "v2"
        >>> Config.from_dict(_data)  # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        pydantic_core._pydantic_core.ValidationError: 1 validation error for Config
        version
          Value error, Version unsupported [type=value_error, input_value='v2', input_type=str]
        ...
        """

        return cls(**data)

    def store(
        self, path: Optional[Path] = None, *, optional_keys: bool = False
    ) -> Optional[str]:
        """
        Write the config file to a specified path.
        If the path is None, then it will be dumped to the stdout.

        The format used is YAML.

        If `optional_keys` is True, then also optional keys will be added to the
        resulting config file.

        >>> _config = Config.from_dict({
        ...     "automsr": {"profiles": [{"email": "foo@gmail.com", "profile": "my-profile"}]},
        ...     "email": {
        ...         "enable": True,
        ...         "recipient": "foo@gmail.com",
        ...         "sender": "bar@gmail.com",
        ...         "sender_password": "abcd",
        ...     },
        ...     "selenium": {
        ...         "profiles_root": "profiles_root",
        ...         "chromedriver_path": "chromedriver",
        ...     }
        ... })

        Dump to stdout
        >>> _config.store()
        '---\\nversion: v1\\nautomsr:\\n  profiles:\\n  - email: foo@gmail.com\\n    profile: my-profile\\n    skip: false\\nselenium:\\n  profiles_root: profiles_root\\n  chromedriver_path: chromedriver\\nemail:\\n  enable: true\\n  recipient: foo@gmail.com\\n  sender: bar@gmail.com\\n  sender_password: abcd\\n'

        Dump to a file
        >>> import tempfile
        >>> f = tempfile.NamedTemporaryFile(mode="w", delete=False)
        >>> _path = Path(f.name)
        >>> _config.store(path=_path)
        >>> f.close()
        >>> assert _path.is_file()
        >>> assert Config.from_yaml(_path) is not None
        >>> _path.unlink()
        """  # noqa: E501

        data = self.get_dict(optional_keys=optional_keys)
        output: Optional[str] = yaml.dump(
            data=data,
            stream=path.open(mode="w") if path is not None else None,
            indent=2,
            sort_keys=False,
            explicit_start=True,
        )
        return output

    def get_dict(self, optional_keys: bool = False) -> Dict[str, Any]:
        """
        Get a dictionary of the current config.

        If `optional_keys` is False, non required keys
        will be filtered out from the final config.
        """

        data = self.model_dump()
        if isinstance(data["email"]["sender_password"], SecretStr):
            data["email"]["sender_password"] = data["email"][
                "sender_password"
            ].get_secret_value()
        if optional_keys:
            return data

        version = data["version"]
        automsr_data = dict(profiles=data["automsr"]["profiles"])
        selenium_data = dict(
            profiles_root=str(data["selenium"]["profiles_root"]),
            chromedriver_path=str(data["selenium"]["chromedriver_path"]),
        )
        email_data = dict(
            enable=data["email"]["enable"],
            # also the following keys are optional,
            # however they are mandatory if `email/enable` is true
            recipient=data["email"]["recipient"],
            sender=data["email"]["sender"],
            sender_password=data["email"]["sender_password"],
        )

        return {
            "version": version,
            "automsr": automsr_data,
            "selenium": selenium_data,
            "email": email_data,
        }


if __name__ == "__main__":
    import doctest

    failures, _all_tests = doctest.testmod()
    assert not failures



import json
import logging
import os
import sqlite3
import sys
from enum import Enum, auto
from pathlib import Path
from typing import ClassVar, Dict, List, Optional, Set, Tuple

from attr import field
from attrs import define

from automsr.config import validate_email

logger = logging.getLogger(__name__)

ENV_HOME = os.environ.get("HOME")
ENV_LOCALAPPDATA = os.environ.get("LOCALAPPDATA")


class ChromeVariant(Enum):
    """
    Chrome variants available
    """

    CHROME = auto()
    CHROME_BETA = auto()
    CHROME_CANARY = auto()
    CHROMIUM = auto()


class OutputFormat(Enum):
    """
    Types of output formats available for printing found Chrome profiles o stdout.
    """

    LIST = "list"
    JSON = "json"
    PRETTY_JSON = "pretty-json"


def get_platform() -> str:
    """
    Returns a simplified version of the system platform in use.
    """

    retval = "linux"  # assumes linux env if not win32 or darwin

    if sys.platform.startswith("win32"):
        retval = "windows"
    elif sys.platform.startswith("darwin"):
        retval = "macOS"

    return retval


@define
class ChromeProfile:
    displayed_name: str
    path: Path

    RESERVED_NAMES = ["System Profile", "Guest Profile"]

    def as_dict(self) -> Dict[str, str]:
        return dict(displayed_name=self.displayed_name, path=str(self.path.resolve()))

    @classmethod
    def from_directory(cls, path: Path) -> Optional["ChromeProfile"]:
        """
        Returns a Chrome Profile based on a directory.

        Returns `None` if the provided `path` is not a valid Chrome Profile.
        """

        # If the provided path is not an existing directory, returns None
        if not path.is_dir():
            return None

        # If the directory name is one of the reserved name for Google profiles
        # not matching any user profile, returns None
        if path.name in cls.RESERVED_NAMES:
            return None

        # If this file doesn't exist, it's not a profile
        preferences_path = path / "Preferences"
        if not preferences_path.is_file():
            return None

        preferences_dict = json.load(open(preferences_path, encoding="utf-8"))
        displayed_name = preferences_dict.get("profile", {}).get("name", "")
        return cls(displayed_name=displayed_name, path=path)

    def get_email(self) -> Optional[str]:
        """
        Try to get the Outlook email address from the profile data.
        """

        allowed_domains = {"outlook", "live", "hotmail", "msn"}

        @define(order=True, frozen=True)
        class Record:
            email: str = field(order=False)
            timestamp: int

            @classmethod
            def from_row(cls, row: Tuple[str, str]) -> Optional["Record"]:
                """
                Parse a row obtained from the Login Data database of Chrome.

                If the row is not compatible with our criteria, returns None.
                """

                assert len(row) == 2
                email_value: str = row[0]
                timestamp_value: int = int(row[1])

                if not validate_email(email_value, raise_on_error=False):
                    return None

                domain = email_value.split("@")[1].split(".")[0]
                if domain not in allowed_domains:
                    return None

                return cls(email=email_value, timestamp=timestamp_value)

        # TODO check if this path is valid for every OS
        login_database: Path = self.path / "Login Data"
        if not login_database.is_file():
            logger.debug("No login database found: %s", login_database)
            return None

        with sqlite3.connect(login_database) as conn:
            cur = conn.execute(
                """\
                select t.username_value, t.date_last_used
                from main.logins t
                where t.username_value <> ''
                and t.origin_url like '%live.com%';"""
            )
            all_rows: List[Tuple[str, str]] = cur.fetchall()

        valid_records: List[Optional[Record]] = [
            Record.from_row(row=row) for row in all_rows
        ]
        valid_non_null_records: List[Record] = [
            record for record in valid_records if record is not None
        ]
        unique_emails: Set[str] = {record.email for record in valid_non_null_records}
        logger.debug("Outlook emails found: %s", unique_emails)

        if not unique_emails:
            logger.debug("No Outlook email found!")
            return None
        elif len(unique_emails) > 1:
            logger.debug(
                "More than one Outlook email found! Will return the latest email used."
            )
            latest_record: Record = max(valid_non_null_records)
            return latest_record.email
        else:
            logger.debug("Found only one Outlook email.")
            return unique_emails.pop()


@define
class ProfilesExecutor:
    chrome_variant: ChromeVariant = ChromeVariant.CHROME
    profiles_root_path: Optional[Path] = None

    # source: https://chromium.googlesource.com/chromium/src/+/master/docs/user_data_dir.md#default-location
    CHROME_DEFAULT_PROFILES_LOCATIONS: ClassVar[
        Dict[str, Dict[ChromeVariant, Path]]
    ] = {
        "macOS": {
            ChromeVariant.CHROME: Path(
                f"{ENV_HOME}/Library/Application Support/Google/Chrome"
            ),
            ChromeVariant.CHROME_BETA: Path(
                f"{ENV_HOME}/Library/Application Support/Google/Chrome Beta"
            ),
            ChromeVariant.CHROME_CANARY: Path(
                f"{ENV_HOME}/Library/Application Support/Google/Chrome Canary"
            ),
            ChromeVariant.CHROMIUM: Path(
                f"{ENV_HOME}/Library/Application Support/Chromium"
            ),
        },
        "windows": {
            ChromeVariant.CHROME: Path(
                f"{ENV_LOCALAPPDATA}\\Google\\Chrome\\User Data"
            ),
            ChromeVariant.CHROME_BETA: Path(
                f"{ENV_LOCALAPPDATA}\\Google\\Chrome Beta\\User Data"
            ),
            ChromeVariant.CHROME_CANARY: Path(
                f"{ENV_LOCALAPPDATA}\\Google\\Chrome SxS\\User Data"
            ),
            ChromeVariant.CHROMIUM: Path(f"{ENV_LOCALAPPDATA}\\Chromium\\User Data"),
        },
        "linux": {
            ChromeVariant.CHROME: Path(f"{ENV_HOME}/.config/google-chrome"),
            ChromeVariant.CHROME_BETA: Path(f"{ENV_HOME}/.config/google-chrome-beta"),
            ChromeVariant.CHROME_CANARY: Path(
                f"{ENV_HOME}/.config/google-chrome-unstable"
            ),
            ChromeVariant.CHROMIUM: Path(f"{ENV_HOME}/.config/chromium"),
        },
    }

    def get_profiles_root_path(self) -> Path:
        """
        Based on current state, get the root path of Chrome profiles.
        """

        if self.profiles_root_path is not None:
            logger.info(
                "Profiles root path was manually set to: %s", self.profiles_root_path
            )
            return self.profiles_root_path

        platform = get_platform()
        logger.debug("Platform to use: %s", platform)
        logger.debug("Chrome variant to use: %s", self.chrome_variant)

        profiles_root_path = self.CHROME_DEFAULT_PROFILES_LOCATIONS[platform][
            self.chrome_variant
        ]
        logger.info(
            "Profiles root path was automatically found to be: %s", profiles_root_path
        )
        return profiles_root_path

    def get_profiles(self) -> List[ChromeProfile]:
        """
        Return the list of Chrome profiles found.
        """

        root_path = self.get_profiles_root_path()

        # Safety check
        if not root_path.exists():
            raise FileNotFoundError(root_path)

        chrome_profiles: List[ChromeProfile] = []

        for directory in root_path.iterdir():
            chrome_profile = ChromeProfile.from_directory(path=directory)
            if chrome_profile is None:
                continue

            logger.info("Found Chrome profile: %s", chrome_profile)
            chrome_profiles.append(chrome_profile)

        return chrome_profiles

    def print_profiles(self, output_format: OutputFormat = OutputFormat.LIST) -> None:
        """
        Get the list of Chrome profiles found, then print it to stdout.

        It's possible to specify a type for printing the values obtained to stdout.
        """

        profiles: List[ChromeProfile] = self.get_profiles()

        if output_format is OutputFormat.LIST:
            data = "\n".join([str(profile) for profile in profiles])

        elif output_format is OutputFormat.JSON:
            data = json.dumps([profile.as_dict() for profile in profiles])

        elif output_format is OutputFormat.PRETTY_JSON:
            data = json.dumps([profile.as_dict() for profile in profiles], indent=4)

        else:
            raise ValueError(output_format)

        data += "\n"  # add manually a final newline
        sys.stdout.write(data)



import logging
import sys
from pathlib import Path
from typing import Any, List, Optional

import questionary
from attr import asdict, define, field
from prompt_toolkit.shortcuts import CompleteStyle
from questionary import Choice, Style

from automsr.browser.profile import ChromeProfile, ChromeVariant, ProfilesExecutor
from automsr.config import Config, validate_email

logger = logging.getLogger(__name__)

# Defaults
DEFAULT_CONFIG_PATH = Path("config.yaml")

# Styles to use with `questionary` prompts
default_style = Style(
    [
        ("highlighted", "fg:#673ab7 bold"),
        ("qmark", "fg:#673ab7 bold"),
        ("question", "green bold"),
        ("answer", "fg:yellow bold"),
    ]
)

error_style = Style(
    [
        ("question", "red bold"),
        ("answer", "fg:yellow bold"),
    ]
)


def handle_null_response(response: Optional[Any]) -> None:
    """
    Handle null responses provided by the user.

    This is related to how `questionary` handles keyboards interrupts,
    both in a safe or unsafe way.
    We are always using the safe way in the execution flow.

    Docs: https://questionary.readthedocs.io/en/stable/pages/advanced.html#keyboard-interrupts
    """

    # the response will be None if the user cancelled the job, e.g., with CTRL+C
    if response is None:
        sys.exit(1)


def fix_unhandled_exception_in_event_loop() -> None:
    """
    This workaround is needed for an issue with the Python event loop
    and python-prompt-toolkit v3.

    The issue is the following:
        "Exception [WinError 995] The I/O operation has been aborted
        because of either a thread exit or an application request

        Press ENTER to continue..."

    Issue: https://github.com/prompt-toolkit/python-prompt-toolkit/issues/1023
    Workaround: https://github.com/xonsh/xonsh/issues/3430
    """

    def enable_selector_event_loop():
        try:
            # These instructions from
            # https://docs.python.org/3/library/asyncio-eventloop.html?highlight=proactor#asyncio.SelectorEventLoop
            import asyncio
            import selectors

            selector = selectors.SelectSelector()
            loop = asyncio.SelectorEventLoop(selector)
            asyncio.set_event_loop(loop)
        except:  # noqa: E722
            pass

    enable_selector_event_loop()


@define(kw_only=True)
class InitExecutor:
    # Selenium config
    chrome_variant: Optional[ChromeVariant] = None
    profiles_root: Optional[Path] = None
    chromedriver_path: Optional[Path] = None

    # Automsr config
    profiles: Optional[List[ChromeProfile]] = None

    # Email config
    email_enabled: Optional[bool] = None
    email_recipient: Optional[str] = None
    email_sender: Optional[str] = None
    email_sender_password: Optional[str] = field(
        default=None, repr=lambda password: "***"
    )

    # Generic config
    config_path: Optional[Path] = None

    def execute(self) -> None:
        """
        Main functionality to call; will prompt the user how to proceed
        in order to generate a config file automatically.
        """

        # This is needed for python-prompt-toolkit
        fix_unhandled_exception_in_event_loop()

        self.check_if_interactive_shell_is_needed()

        if self.chrome_variant is None:
            self.chrome_variant = self.get_chrome_variant()

        if self.profiles_root is None:
            self.profiles_root = self.get_chrome_profiles_root()

        if self.profiles is None:
            self.profiles = self.get_chrome_profiles()

        if self.chromedriver_path is None:
            self.chromedriver_path = self.get_chromedriver_path()

        if self.config_path is None:
            self.config_path = self.get_config_path()

        if self.email_enabled is None:
            self.email_enabled = self.get_email_enabled()

        if self.email_enabled:
            if self.email_recipient is None:
                self.email_recipient = self.get_email_recipient()
            if self.email_sender is None:
                self.email_sender = self.get_email_sender()
            if self.email_sender_password is None:
                self.email_sender_password = self.get_email_sender_password()

        config = self.get_config()
        self.store_config(config=config)

    def get_config(self) -> Config:
        """
        Get a config representation based on the status of the object.
        """

        assert self.profiles is not None
        dummy_email = "must-change-this-with-a-correct-email-address@outlook.com"
        profiles_obj = [
            {"email": p.get_email() or dummy_email, "profile": p.path.stem}
            for p in self.profiles
        ]
        if any(profile["email"] == dummy_email for profile in profiles_obj):
            questionary.print(
                text=(
                    "At least one email address was not found automatically. "
                    "Please add it yourself in the config file."
                ),
                style="red bold",
            )
        automsr_obj = {"profiles": profiles_obj}

        assert self.profiles_root is not None
        assert self.chromedriver_path is not None
        selenium_obj = {
            "profiles_root": str(self.profiles_root),
            "chromedriver_path": str(self.chromedriver_path),
        }

        assert self.email_enabled is not None

        if self.email_enabled:
            assert self.email_recipient is not None
            assert self.email_sender is not None
            assert self.email_sender_password is not None
        else:
            logger.debug(
                "Not performing email attributes' check for None, since email sending is disabled."
            )

        email_obj = {
            "enable": self.email_enabled,
            "recipient": self.email_recipient,
            "sender": self.email_sender,
            "sender_password": self.email_sender_password,
        }

        config = Config.from_dict(
            data={
                "automsr": automsr_obj,
                "selenium": selenium_obj,
                "email": email_obj,
            }
        )

        return config

    def store_config(self, config: Config) -> None:
        """
        Store a config object to disk based on the state of this object.
        """

        assert self.config_path is not None
        config.store(path=self.config_path)

    def check_if_interactive_shell_is_needed(self) -> None:
        """
        Check if an interactive shell is needed, based on the state of the object.

        Raise if is needed and the current shell is not interactive.
        """

        members = asdict(self)

        acceptable_null_members = {"profiles_root"}
        for member in acceptable_null_members:
            members.pop(member)

        if any(value is None for value in members.values()):
            self.assert_interactive_shell()

    @staticmethod
    def assert_interactive_shell() -> None:
        """
        Check if the current shell is interactive.

        Raise a `RuntimeError` if this is not True.
        """

        if not sys.stdout.isatty():
            logger.error("Current shell is not interactive!")
            logger.warning(
                "To enable this command to be executed, please provide non-null values for every class member."
            )
            raise RuntimeError("Current shell is not interactive!")
        else:
            logger.debug("Current shell is interactive.")

    @staticmethod
    def get_chrome_variant() -> ChromeVariant:
        """
        Get the Chrome variant after asking the user to select it from a list.
        """

        chrome_variants: List[str] = [
            variant.name.replace("_", " ").capitalize() for variant in ChromeVariant
        ]

        chrome_variant: Optional[str] = questionary.select(
            message="Which variant of Chrome are you using?",
            choices=chrome_variants,
            use_indicator=True,
            style=default_style,
        ).ask()

        handle_null_response(chrome_variant)
        assert chrome_variant is not None

        variant = list(ChromeVariant)[chrome_variants.index(chrome_variant)]
        logger.info("Variant selected: %s", variant)
        return variant

    def _get_profiles_executor(self) -> ProfilesExecutor:
        """
        Get a Profiles Executor based on the status of the object.
        """

        if self.chrome_variant is not None:
            return ProfilesExecutor(
                chrome_variant=self.chrome_variant,
                profiles_root_path=self.profiles_root,
            )
        else:
            return ProfilesExecutor(profiles_root_path=self.profiles_root)

    def get_chrome_profiles_root(self) -> Path:
        """
        Get the current Chrome profiles root.
        """

        profiles_executor = self._get_profiles_executor()
        return profiles_executor.get_profiles_root_path()

    def get_chrome_profiles(self) -> List[ChromeProfile]:
        """
        Set the Chrome profiles based on the selected Chrome variant and the user prompt.
        """

        profiles_executor = self._get_profiles_executor()
        try:
            profiles = profiles_executor.get_profiles()
        except FileNotFoundError as e:
            logger.error(
                "Profiles root directory was not found! "
                "Are you sure that you selected the correct Chrome variant?"
            )
            raise e

        if not profiles:
            logger.error(
                "No Chrome profile found! Create at least one, then try again."
            )
            raise RuntimeError("No Chrome profile found!")

        profiles_as_str = [str(profile) for profile in profiles]
        chosen_profiles: Optional[List[str]] = questionary.checkbox(
            message="Found the following profiles. Select the ones you intend to use.",
            choices=[
                Choice(title=profile, checked=False) for profile in profiles_as_str
            ],
            validate=lambda choices: len(choices) > 0 or "Select at least one profile",
            style=default_style,
        ).ask()

        handle_null_response(chosen_profiles)
        assert chosen_profiles is not None

        profiles_to_use: List[ChromeProfile] = []
        for profile_str in chosen_profiles:
            index: int = profiles_as_str.index(profile_str)
            profile = profiles[index]
            profiles_to_use.append(profile)

        logger.info("Chosen profiles: %s", profiles_to_use)
        return profiles_to_use

    @staticmethod
    def get_chromedriver_path() -> Path:
        """
        Get the Chromedriver path after prompting the user where to find it.
        """

        path_str: Optional[str] = questionary.path(
            "What's the path to the Chromedriver executable?",
            validate=lambda v: Path(v).is_file() or "File not found",
            complete_style=CompleteStyle.READLINE_LIKE,
            style=default_style,
        ).ask()
        handle_null_response(path_str)
        assert path_str is not None
        return Path(path_str)

    @staticmethod
    def get_config_path() -> Path:
        """
        Get the path to the config file.
        """

        while True:
            path_str: Optional[str] = questionary.path(
                "Output path for the YAML Config file. Leave empty for the default: ./config.yaml",
                complete_style=CompleteStyle.READLINE_LIKE,
                validate=lambda v: v == ""
                or not Path(v).is_dir()
                or "Path invalid or directory",
                style=default_style,
            ).ask()
            handle_null_response(path_str)
            assert path_str is not None
            if not path_str:
                path = DEFAULT_CONFIG_PATH
            else:
                path = Path(path_str)
            logger.info("Path provided: %s", path)

            if path.is_file():
                will_overwrite: Optional[bool] = questionary.confirm(
                    message="The path you provided already exists! Do you want to overwrite it?",
                    default=False,
                    style=error_style,
                ).ask()
                handle_null_response(will_overwrite)
                assert will_overwrite is not None
                if will_overwrite:
                    break
            else:
                break

        logger.info("Path that will be used for the config: %s", path)
        return path

    @staticmethod
    def get_email_enabled() -> bool:
        """
        Get a boolean representing if the email sending should be enabled or not.
        """

        email_enabled: Optional[bool] = questionary.confirm(
            message="Do you want to enable the sending of a status email after the execution?",
            default=True,
            style=default_style,
        ).ask()
        handle_null_response(email_enabled)
        assert email_enabled is not None
        return email_enabled

    @staticmethod
    def get_email_recipient() -> str:
        """
        Get the recipient for the status email.
        """

        recipient: Optional[str] = questionary.text(
            message="What is the recipient of the status email?",
            validate=lambda x: True
            if validate_email(x, raise_on_error=False) is not None
            else "Invalid email!",
            style=default_style,
        ).ask()
        handle_null_response(recipient)
        assert recipient is not None
        return recipient

    @staticmethod
    def get_email_sender() -> str:
        """
        Get the sender for the status email.
        """

        sender: Optional[str] = questionary.text(
            message="What is the sender of the status email?",
            validate=lambda x: True
            if validate_email(x, raise_on_error=False) is not None
            else "Invalid email!",
            style=default_style,
        ).ask()
        handle_null_response(sender)
        assert sender is not None
        return sender

    @staticmethod
    def get_email_sender_password() -> str:
        """
        Get the sender's password for the status email.
        """

        password: Optional[str] = questionary.password(
            message=(
                "What is the password of the sender email? "
                "Please use an app password if you activated 2FA on this account."
            ),
            validate=lambda x: bool(x.strip()) or "Empty password!",
            style=default_style,
        ).ask()
        handle_null_response(password)
        assert password is not None
        return password
